"""
norm.py
A galaxy tool to normalise CEL files using RMA
This python script uses a subprocess to call an R script stored
here as a string called rprog

"""
import sys, string, os, tempfile, zipfile, shutil, subprocess, re

rprog="""
# Function to retrieve custom REFSEQ CDF provided Affymetrix Chip name
chiptorefseq <- function (chip)
{
    ctable <- read.table("customREFSEQ", sep = "\t", header = TRUE)
    results <- as.vector(ctable$Custom[which(ctable$Chip == chip )])    
    return(results)   
}

# Function to retrieve custom ENTREZG CDF provided Affymetrix Chip name
chiptoentrezg <- function (chip)
{
    ctable <- read.table("customENTREZG", sep = "\t", header = TRUE)
    results <- as.vector(ctable$Custom[which(ctable$Chip == chip )])    
    return(results)   
}
rnorm <- function(tdir,logfpath,title,method)
#  rcall = "%s('%s','%s','%s','%s')" % (rnorm,tdir,logfpath,title,method)

{
sink(file=file(logfpath, "at"), type="message")
sink(file=file(logfpath, "at"), type="output")
newfiles = c()
newnames = c()
newtypes = c()
setwd(tdir)
isCel <- 0
isXys <- 0
fn <- dir(pattern="[Cc][Ee][Ll]")
if (length(fn) > 0) {
    isCel <- 1
} else {
    fn <- dir(pattern="[Xx][Yy][Ss]")
    if (length(fn) > 0) {
        isXys <- 1
    } else {
        return(list(newfiles=c(),newnames=c(),newtypes=c()))
    }
}
pd <- dir(pattern="[Tt][Xx][Tt]")

library(oligo) 
library(Biobase)
library(geneplotter) 
library(RColorBrewer) 
library(genefilter)
palette(brewer.pal(8, "Dark2"))

if (isXys) {
    library(maqcExpression4plex)
    library(limma)
}

rawname = paste(title, method, sep="_")
rawname = paste(rawname, "Raw_Plots.pdf", sep="_")
rawfname = file.path(tdir, rawname)
normname = paste(title, method, sep="_")
normname = paste(normname, "Normalised_Plots.pdf", sep="_")
normfname = file.path(tdir, normname)

if (isCel) {
    esetname = paste(title, method, sep="_")
    esetname = paste(esetname, "eset", sep=".")
    esetfname = file.path(tdir, esetname)

    esetvaluesname = paste(title, method, "refseq_values", sep="_")
    esetvaluesname = paste(esetvaluesname, "txt", sep=".")
    esetvaluesfname = file.path(tdir, esetvaluesname)

    esetvaluesname2 = paste(title, method, "entrez_values", sep="_")
    esetvaluesname2 = paste(esetvaluesname2, "txt", sep=".")
    esetvaluesfname2 = file.path(tdir, esetvaluesname2) 

    #Inserted with new ProbeSetID type  
    esetvaluesname3 = paste(title, method, "ProbeSetID_values", sep="_")
    esetvaluesname3 = paste(esetvaluesname3, "txt", sep=".")
    esetvaluesfname3 = file.path(tdir, esetvaluesname3)   
    #########

    newnames = c(newnames, rawname, normname, esetname, esetvaluesname, esetvaluesname2)
    newfiles = c(newfiles, rawfname, normfname, esetfname, esetvaluesfname, esetvaluesfname2)

    #Inserted with new ProbeSetID type
    newnames = c(newnames, rawname, normname, esetname, esetvaluesname, esetvaluesname3)
    newfiles = c(newfiles, rawfname, normfname, esetfname, esetvaluesfname, esetvaluesfname3)
    #########


    newtypes = c(newtypes, 'pdf', 'pdf', 'eset', 'txt', 'txt') 
} else {
    esetname = paste(title, method, sep="_")
    esetname = paste(esetname, "eset", sep=".")
    esetfname = file.path(tdir, esetname)
    esetvaluesname = paste(title, method, "values", sep="_")
    esetvaluesname = paste(esetvaluesname, "txt", sep=".")
    esetvaluesfname = file.path(tdir, esetvaluesname)
    newnames = c(newnames, rawname, normname, esetname, esetvaluesname)
    newfiles = c(newfiles, rawfname, normfname, esetfname, esetvaluesfname)
    newtypes = c(newtypes, 'pdf', 'pdf', 'eset', 'txt') 
}

p <- read.delim(pd, header=TRUE)
if (isCel) {
    library(affy) 
    testRead = ReadAffy(filenames = fn[1])
    db = annotation(testRead)
    refseqCdf = chiptorefseq(db)
    entrezCdf = chiptoentrezg(db)

    rawRefSeq <- ReadAffy(filenames = fn, phenoData = p, cdfname = refseqCdf)
    rawEntrez <- ReadAffy(filenames = fn, phenoData = p, cdfname = entrezCdf)
    raw <- ReadAffy(filenames = fn, phenoData = p)

    ### Cut out .cel from filename to make readable
    names_length <- nchar(sampleNames(rawRefSeq))
    sampleNames(rawRefSeq) <- substr(sampleNames(rawRefSeq),start=1, stop = (names_length-4)) #cut .cel from name
    ### Cut out .cel from filename to make readable
    names_length <- nchar(sampleNames(rawEntrez))
    sampleNames(rawEntrez) <- substr(sampleNames(rawEntrez),start=1, stop = (names_length-4)) #cut .cel from name
} else {
    basename(fn)
    raw <- read.xysfiles(fn, phenoData = p)
    phenoData(raw) <- read.AnnotatedDataFrame(pd)
    db <- annotation(raw)

    ### Cut out .cel from filename to make readable
    names_length <- nchar(sampleNames(raw))
    sampleNames(raw) <- substr(sampleNames(raw),start=1, stop = (names_length-4)) #cut .cel from name
}


### Download annotation for a chip we've not seen before
x <-(sub(' ','',(paste(annotation(raw), ".db"))))
if (! require(x, character.only=T)) {
    source("http://bioconductor.org/biocLite.R")
    biocLite(x)
    library(x, character.only=T, logical.return=T)
}

require(graphics)
par(mfrow=c(2,1))
pdf(file=rawfname, paper='a4')
boxplot(raw, las=2, cex.axis=0.5, main="Raw box plot", col=p$Group)
hist(raw, las=2, cex.axis=0.5, main="Raw histogram", col=p$Group)
dev.off()

if (method == 'rma') {
  if (isCel) {
    esetRefSeq <- rma(rawRefSeq)
    esetEntrez <- rma(rawEntrez)
    eset <- rma(raw)

    #varCut = 0.5
    #filt_bcrneg = nsFilter(esetEntrez, require.entrez=TRUE,require.GOBP=TRUE, remove.dupEntrez=TRUE, var.func=IQR, var.cutoff=varCut,feature.exclude="^AFFX")
    #esetEntrez <- exprs(filt_bcrneg$esetEntrez)
  } else {
    eset <- rma(raw)
  }
}
if (method == 'justrma') {
  if (isCel) {
    esetRefSeq <- justRMA(filenames=fn, phenoData = p, cdfname = refseqCdf)
    esetEntrez <- justRMA(filenames=fn, phenoData = p, cdfname = entrezCdf)
    eset <- justRMA(filenames=fn, phenoData = p)
  } else {
    eset <- justRMA(filenames=fn, phenoData = p)
  }
}
if (method == 'gcrma') {
  library(gcrma)
  if (isCel) {
    esetRefSeq <- gcrma(rawRefSeq)
    esetEntrez <- gcrma(rawEntrez)
    eset <- gcrma(raw)
  } else {
    eset <- gcrma(raw)
  }
}
if (method == 'mas5') {
  if (isCel) {
    esetRefSeq <- mas5(rawRefSeq)
    esetEntrez <- mas5(rawEntrez)
    eset <- mas5(raw)
  } else {
    eset <- mas5(raw)
  }
}

if (isCel) {
    save(eset, file=esetfname)
    ProbeSetID <- exprs(eset)
    refSeqEsetValues <- exprs(esetRefSeq)
    entrezEsetValues <- exprs(esetEntrez)

    rownames(refSeqEsetValues) <- strsplit(as.vector(rownames(refSeqEsetValues)), '_at')
    rownames(entrezEsetValues) <- strsplit(as.vector(rownames(entrezEsetValues)), '_at')

    write.table(refSeqEsetValues, file=esetvaluesfname, sep = "\t", quote = FALSE, col.names = NA)
    write.table(entrezEsetValues, file=esetvaluesfname2, sep = "\t", quote = FALSE, col.names = NA)
    write.table(ProbeSetID, file=esetvaluesfname3, sep="\t", quote =FALSE, col.names =NA)
} else {
    save(eset, file=esetfname)
    esetvalues <- exprs(eset)
    write.table(esetvalues, file=esetvaluesfname)
}

par(mfrow=c(2,1))
pdf(file=normfname, paper='a4')
boxplot(eset, las=2, cex.axis=0.5, main="Normalised box plot", col=p$Group)
hist(eset, las=2, cex.axis=0.5, main="Normalised histogram", col=p$Group)
dev.off()

cat(paste("DB:", db, "\\n"))

sink()
return(list(newfiles=newfiles,newnames=newnames,newtypes=newtypes,db=db))
}
"""

mng = '### makenewgalaxy' # used by exec_after_process hook to parse out new file paths/names
tmp = '### tmp' # used by exec_after_process hook to clean up tmp directory
dbString = '### db'

def unzip(zipFileName, tmpDir, log):
    log.write("Creating AffyBatch object from CEL files\n")
    #modified to incorporate folders inside zip file
    zipFile = zipfile.ZipFile(zipFileName, "r")

    for zipInfo in zipFile.infolist():
        fileName = zipInfo.filename.strip()

        fileBaseName = os.path.basename(fileName)
        if(fileBaseName=="" or fileBaseName.startswith=="." or fileName.startswith("__MACOSX")):
            # ignore folder name, hidden file in *nix, or extra resource forks by Max OSX ZIP software.
            continue

        bytes = zipFile.read(fileName)
        
        if(fileName.rfind("/")!=""):
            folderArr = fileName.split("/")
            fileName = folderArr[len(folderArr)-1]            

        fileName = tmpDir + "/" + fileName

        log.write("fileName:"+fileName)

        f = open(fileName, "wb")
        f.write(bytes)
        f.close()

    zipFile.close()

def getPheno(tDir, log):
    log.write("Constructing pheno info\n")

    files = os.listdir(tDir)
    phenoFileName = None
    for (f) in files:
        if (os.path.splitext(f)[1].lower() == '.txt'):
            phenoFileName = f

    if (phenoFileName <> None):
        f1 = os.path.join(tDir, phenoFileName)
        f2 = os.path.join(tDir, phenoFileName + ".2")
        phenoFile = file(f1, 'rU')
        pheno = phenoFile.readlines()

        #Find Group Column Position
        headerItems = pheno[0].strip().split('\t')
        geoupColumnMatch = re.compile('^Group$', re.I)

        i=0
        groupColumnPosn=1
        for (hdr) in headerItems:
            if(geoupColumnMatch.match(hdr)):
                groupColumnPosn = i
                break
            i=i+1

        cleanPheno = file(f2, 'w')
        sb = ""
        tokens = [x.strip().split('\t') for x in pheno]
        for phenoLine in tokens:
            tabDedup = phenoLine[0].strip() + '\t' + phenoLine[groupColumnPosn].strip() + '\n'
            sb = sb + '### pheno\t' + tabDedup
            cleanPheno.write(tabDedup)

        log.write(sb + '\n')

        cleanPheno.close()
        phenoFile.close()

        os.rename(f2, f1)
    else:
        log.write('No pheno data!\n')
            
def main():
    """called as 
<command interpreter="python">
rnorm.py '$i' '$title' '$logmeta' '$method'
</command>
    """
    nparm = 4
    appname = sys.argv[0]
    if (len(sys.argv) < nparm):
        print '%s needs %d parameters - given %d as %s' % (appname,nparm,len(sys.argv),';'.join(sys.argv))
        print 'usage: %s zip_file title out_file method' % (appname)
        sys.exit(1)
        # short command line error
    infpath = sys.argv[1].strip()
    title = sys.argv[2].strip()
    logfname = sys.argv[3].strip()
    method = sys.argv[4].strip()
    runInPath = appname.rsplit('/', 1)[0]
    replace = string.whitespace + string.punctuation
    ttab = string.maketrans(replace,'_'*len(replace))
    title = string.translate(title,ttab)
    tdir = tempfile.mkdtemp(prefix=title)
    logf = file(logfname,'w')
    logf.write('# %s - running' % (appname))
    logf.write('# Got parameters %s\n' % ' '.join(sys.argv))
    unzip(infpath, tdir, logf)
    getPheno(tdir, logf)
    logf.close()

    # copy custom cdf for the run
    shutil.copy(runInPath + "/customENTREZG", tdir)
    shutil.copy(runInPath + "/customREFSEQ", tdir)

    rprogname = 'rnorm'
    rcall = "%s('%s','%s','%s','%s')" % (rprogname,tdir,logfname,title,method)
    print '##rcall=',rcall

    p = subprocess.Popen("R --vanilla", shell=True, executable="/bin/bash", stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    (stdout, stderr) = p.communicate(rprog + rcall)
    makenew = makeFilenames(tdir, title, method, logfname)
    newfiles = makenew[0]
    newnames = makenew[1]
    newtypes = makenew[2]
    db = makenew[3][0]

    logf = file(logfname,'a')
    s = 'R code and call\n%s\n\n\nCalled as\n%s\n' % (rprog, rcall)
    logf.write(s)
    try:
        outlist = ['%s\t%s\t%s\t%s' % (mng,newfiles[i],newnames[i],newtypes[i]) for i in range(len(newfiles))]
    except:
        outlist = ['%s\t%s\t%s\t%s' % (mng,newfiles,newnames,newtypes),] # only 1
    tmpString = '\n%s\t%s\n' % (tmp, tdir)
    logf.write(tmpString)
    logf.write('%s\t%s\n' % (dbString, db))
    logf.write('\n'.join(outlist))
    logf.write('\n')
    logf.close()

def makeFilenames(tdir, title, method, logfname):
    files = os.listdir(tdir)
    files = [x.lower() for x in files]
    cels = [x for x in files if x.endswith(".cel")]
    xys = [x for x in files if x.endswith(".xys")]
    isCel = len(cels) > 0
    isXys = len(xys) > 0

    if (not isCel and not isXys):
        return [[], [], []]

    rawname = title + "_" + method + "_Raw_Plots.pdf"
    rawfname = os.path.join(tdir, rawname)
    normname = title + "_" + method + "_Normalised_Plots.pdf"
    normfname = os.path.join(tdir, normname)
    esetname = title + "_" + method + ".eset"
    esetfname = os.path.join(tdir, esetname)

    log = open(logfname, "r")
    logs = log.readlines()
    log.close()
    logs = [x for x in logs if x.startswith("DB:")]
    db = [x.split("DB:")[1].strip() for x in logs]

    if (isCel):
        esetvaluename = title + "_" + method + "_refseq_values.txt"
        esetvaluefname = os.path.join(tdir, esetvaluename)
        
        esetvaluename2 = title + "_" + method + "_entrez_values.txt"
        esetvaluefname2 = os.path.join(tdir, esetvaluename2)
	
        esetvaluename3 = title + "_" + method + "_ProbeSetID_values.txt"
	esetvaluefname3 = os.path.join(tdir, esetvaluename3)
        return [[rawfname, normfname, esetfname, esetvaluefname, esetvaluefname2, esetvaluefname3],
                [rawname, normname, esetname, esetvaluename, esetvaluename2, esetvaluename3],
                ['pdf', 'pdf', 'eset', 'txt', 'txt', 'txt']
                , db
               ]
    else:
        esetvaluename = title + "_" + method + "_values.txt"
        esetvaluefname = os.path.join(tdir, esetvaluename)
        return [[rawfname, normfname, esetfname, esetvaluefname],
                [rawname, normname, esetname, esetvaluename],
                ['pdf', 'pdf', 'eset', 'txt']
                , db
               ]

if __name__ == "__main__":
    main()

